use std::time;
use std::net::SocketAddr;
use ether_dream::dac::stream::connect;
use ether_dream::dac::{Playback, Stream};

use crate::conf::EtherDreamConf;
use crate::device::{Device, Status, PlaybackState};
use crate::errors::{LJError, LJResult};
use crate::point::{Color, Point};
use ether_dream::protocol::{DacBroadcast, DacPoint};
use log::{debug, info, warn};

#[warn(dead_code)]
pub struct EtherdreamDevice {
	pub conf: EtherDreamConf,
	dac: DacBroadcast,
	// source_address: SocketAddr,
	stream: Stream,
	// sent_points: u16,
	lack: String,
	last_traced_at: String,
}


impl EtherdreamDevice {
	pub fn new(conf: &EtherDreamConf) -> LJResult<Self> {
		let (dac, _source_address, stream) = EtherdreamDevice::get_dac(conf)?;

		Ok(Self {
			conf: (*conf).clone(),
			dac,
			// source_address,
			stream,
			// sent_points: 0,
			lack: "".to_string(),
			last_traced_at: "1985-04-12T23:20:50.52Z".to_string(),
		})
	}
	pub fn get_dac(conf: &EtherDreamConf) -> LJResult<(DacBroadcast, SocketAddr, Stream)> {
		let ip = &conf.ip;
		let dac_broadcast = ether_dream::recv_dac_broadcasts()?;
		dac_broadcast.set_timeout(Some(time::Duration::new(10, 0)))?;
		info!("Attempting to get DAC broadcast...");
		let broadcast = dac_broadcast
			.take(3)
			.filter_map(|result| {
				match result {
					Err(err) => {
						warn!( "Failed to find a valid DAC via broadcast. Error: {:?}", err);
						info!( "Retrying...");
						None
					}
					Ok((dac, source_addr)) => {
						info!("Valid broadcast, source_addr: {}", source_addr);
						if source_addr.is_ipv6() { return None; }
						if &source_addr.ip().to_string() != ip { return None; }
						Some(Ok((dac, source_addr)))
					}
				}
			})
			.next()
			.expect("Failed to receive broadcast.");
		match broadcast {
			Err(err) => {
				Err(Box::new(LJError::EtherdreamConnectError(err)))
			}
			Ok((dac, source_addr)) => {
				let stream = EtherdreamDevice::get_tcp_stream(&dac, &source_addr)?;
				info!("Finished configuring DAC and TCP stream.");
				Ok((dac, source_addr, stream))
			}
		}
	}

	pub fn get_tcp_stream(dac: &DacBroadcast, source_address: &SocketAddr) -> LJResult<Stream> {
		// Establish the TCP connection.
		let mut stream = connect(dac, source_address.ip())?;
		EtherdreamDevice::prepare_tcp_stream(&mut stream).unwrap();
		Ok(stream)
	}

	fn prepare_tcp_stream(stream: &mut Stream) -> LJResult<()> {

		// Prepare  stream
		match stream
			.queue_commands()
			.prepare_stream()
			.submit() {
			Err(err) => {
				warn!(
					"err occurred when submitting PREPARE_STREAM command and listening for response: {}",
					err
				);
			}
			Ok(_) => {
				info!("Prepared Stream.")
			}
		}

		let control = 0;
		let (u1, u2) = (0, 0);
		let i = 255;
		let point = DacPoint {
			control,
			x: 0,
			y: 0,
			i,
			r: 0,
			g: 0,
			b: 0,
			u1,
			u2,
		};
		let begin_list = vec![point];
		let points_per_second = stream.dac().max_point_rate / 32;
		match stream
			.queue_commands()
			.data(begin_list.into_iter().take(1 as usize))
			.begin(0, points_per_second)
			.submit() {
			Err(err) => {
				warn!(
					"err occurred when submitting first data: {}",
					err
				);
			}
			Ok(_) => {
				info!("Sent first data to Etherdream.")
			}
		}

		Ok(())
	}


	pub fn check_tcp_stream(&mut self) -> LJResult<()> {
		// todo Reinit stream if needed
		// self.stream = EtherdreamDevice::get_tcp_stream(&self.dac, &self.source_address)?
		Ok(())
	}

	// Determine the number of points needed to fill the DAC.
	fn points_capacity(&self) -> u16 {
		// Fixme thread 'main' panicked at 'attempt to subtract with overflow', src/device/etherdream.rs:144:24
		let n_points = self.dac.buffer_capacity as u16 - self.stream.dac().dac.status.buffer_fullness as u16 - 1;
		n_points
	}
}

impl Device for EtherdreamDevice {
	fn status(&mut self) -> Status {
		let _ = self.check_tcp_stream();

		// "a": ACK   "F": Full  "I": invalid. 64 or 35 for no connection.
		let playback_state = match self.stream.dac().dac.status.playback {
			Playback::Idle => PlaybackState::IDLE,
			Playback::Prepared => PlaybackState::PREPARE,
			Playback::Playing => PlaybackState::PLAYING,
		};

		let status = Status {
			last_traced_at: self.last_traced_at.clone(),
			properties: vec!["foo".to_string()],
			playback_state,
			capacity: self.points_capacity(),
			lack: String::from(&self.lack),
		};
		// info!("Dac Status: {:?} ", status );
		// info!("Etherdream Dac {:?} ", self.dac );
		// info!("Stream dac{:?}", self.stream.dac());

		status
	}

	fn draw(&mut self,
	        line: Vec<Point>,
	        _speed: u32,
	) -> LJResult<()> {
		let n_points = self.points_capacity();
		// let n_points = &line.len();
		debug!("Etherdream::device draw Generating {:?} points", n_points);
		return match self.stream
			.queue_commands()
			.data(
				line.into_iter()
					.map(|point| point.into())
					// .take(line.len() as usize)
					.take(n_points as usize )
			)
			.submit() {
			Err(err) => {
				// We should account for
				// 'Broken pipe (os error 32)'
				// Connection reset by peer (os error 104)
				warn!("Draw error: '{}'",err);
				Ok(())
			}
			Ok(_) => {
				debug!("Draw is ok");
				Ok(())
			}
		};
	}

	fn stop(&mut self) -> LJResult<()> {
		info!("Stopping Etherdream device...");
		match self.stream
			.queue_commands()
			.stop()
			.submit()
		{
			Err(err) => {
				warn!("Failed to stop EtherDream device with error {:?}", err);
				Err(Box::new(err))
			}
			Ok(_) => {
				info!("Sucessfully closed EtherDream device.");
				Ok(())
			}
		}
	}

	fn grid(&mut self) -> Vec<Point> {
		let dim_mid = 16000 as f32;
		let dim_max = 32000 as f32;
		let col_min = Color { r: 0, g: 0, b: 0 };
		let col_max = Color { r: 255, g: 255, b: 255 };

		vec![
			Point { x: -dim_max, y: dim_max, color: col_min },
			Point { x: -dim_max, y: dim_max, color: col_max },
			Point { x: dim_max, y: dim_max, color: col_max },
			Point { x: dim_max, y: -dim_max, color: col_max },
			Point { x: -dim_max, y: -dim_max, color: col_max },
			Point { x: -dim_max, y: -dim_mid, color: col_min },
			Point { x: -dim_mid, y: dim_mid, color: col_min },
			Point { x: -dim_mid, y: dim_mid, color: col_max },
			Point { x: dim_mid, y: dim_mid, color: col_max },
			Point { x: dim_mid, y: -dim_mid, color: col_max },
			Point { x: -dim_mid, y: -dim_mid, color: col_max },
			Point { x: -dim_mid, y: -dim_mid, color: col_min },
		]
	}
}
